__author__ = "Jeremy Lainé"
__email__ = "jeremy.laine@m4x.org"
__license__ = "BSD"
__summary__ = "Python wrapper for the ls-qpack QPACK library"
__title__ = "pylsqpack"
__uri__ = "https://github.com/aiortc/pylsqpack"
__version__ = "0.3.7"
